package srtree;

import java.util.*;

/**
 * A cached persistent page file with a Least Recently Used (LRU) caching strategy.<br>
 * (PENDING: Use a sorted list on rankings here, for better performance.)
 * <p>
 * Created: Sun Oct 31 19:22:05 1999
 * <p>
 * @author Hadjieleftheriou Marios
 * @version 1.001
 */
public class CachedPersistentPageFile extends PersistentPageFile {
    private Hashtable cache;
    private int usedSpace = 0;
    private int cacheSize;

    protected long cacheHit = 0;

    public CachedPersistentPageFile(String fileName, int cacheSize) {
	super(fileName);
	this.cacheSize = cacheSize;
	cache = new Hashtable(cacheSize);
    }

    protected void clearCache() {
	cache.clear();
    }

    protected AbstractNode readNode(long page) throws PageFaultError {
	AbstractNode n = (AbstractNode) readFromCache(page);
	if (null != n) {
	    cacheHit++;
	    return n;
	} else {
	    return super.readNode(page);
	}
    }

    protected long writeNode(AbstractNode n) throws PageFaultError {
	long page = super.writeNode(n);
	writeToCache(n, page);
	return page;
    }

    private Object readFromCache(long page) {
	CachedObject c = (CachedObject) cache.get(new Long(page));
	if (c != null) {
	    int rank = c.rank;
	    for (Enumeration e = cache.elements(); e.hasMoreElements();) {
		CachedObject co = (CachedObject) e.nextElement();
		if (co.rank > rank) {
		    co.rank--;
		} else if (co.rank == rank) {
		    co.rank = usedSpace-1;
		}
	    }
	    return c.object;
	} else {
	    return null;
	}
    }

    private void writeToCache(Object o, long page) {
	CachedObject c = (CachedObject) cache.get(new Long(page));

	if (null != c) {
	    // if object already in cache.
	    c.object = o;
	    int rank = c.rank;
	    for (Enumeration e = cache.elements(); e.hasMoreElements();) {
		CachedObject co = (CachedObject) e.nextElement();
		if (co.rank > rank) {
		    co.rank--;
		} else if (co.rank == rank) {
		    co.rank = usedSpace-1;
		}
	    }
	} else if (usedSpace < cacheSize) {
	    // if cache is not full yet.
	    cache.put(new Long(page), new CachedObject(o, page, usedSpace));
	    usedSpace++;
	    return;
	} else {
	    // if cache is full.
	    for (Enumeration e = cache.elements(); e.hasMoreElements();) {
		c = (CachedObject) e.nextElement();
		if (c.rank == 0) {
		    cache.remove(new Long(c.page));
		    break;
		}
	    }

	    for (Enumeration e = cache.elements(); e.hasMoreElements();) {
		c = (CachedObject) e.nextElement();
		c.rank--;
	    }

	    cache.put(new Long(page), new CachedObject(o, page, usedSpace-1));
	}
    }

    class CachedObject {
	int rank = 0;
	long page = -1;
	Object object;

	public CachedObject(Object o, long page, int rank) {
	    this.object = o;
	    this.page = page;
	    this.rank = rank;
	}
    }

    public void dumpFile() {
	super.dumpFile();
	System.out.println("Cache hits: " + cacheHit);
    }
    
} // CachedPersistentPageFile
