package srtree;

/**
 * Internal node of the SRTree. Used to access Leaf nodes, where real data lies.
 * <p>
 * Created: Tue May 18 13:04:25 1999
 * <p>
 * @author Hadjieleftheriou Marios
 * @version 1.001
 */

import java.util.*;

public class Index extends AbstractNode {
    /**
     * Spanning segments are stored into this array.
     */
    protected HyperCube[] segments;

    /**
     * How much space is used up for segment entries. If equal to
     * nodeCapacity * (1 - branchFactor), then segment array is full.
     */
    protected int segmentSpace = 0;

    /**
     * The index of the data node spanned by each segment.
     */
    protected int[] spanned;

    protected Index(SRTree tree, long parent, long pageNumber, int level) {
	super(tree, parent, pageNumber, level);
	int b = (int) (Math.pow(2, level) * tree.getNodeCapacity() * tree.getBranchFactor()) + 1;
	data = new HyperCube[b];
	branches = new long[b];
	usedSpace = 0;

	int s = (int) (Math.pow(2, level) * tree.getNodeCapacity() * (1 - tree.getBranchFactor()));
	segments = new HyperCube[s];
	spanned = new int[s];
	segmentSpace = 0;
    }

    protected AbstractNode chooseNode(HyperCube h) {
	int i;

	for (i = 0; i < usedSpace; i++) {
	    if (h.spansOver(data[i], 0)) {
		// ok, the hypercube spans an entry, so insert it here.
		return this;
	    }
	}

	i = findLeastEnlargement(h);
	return ((AbstractNode) getChild(i)).chooseNode(h);
    }

    protected Leaf findLeaf(HyperCube h) {
	for (int i = 0; i < usedSpace; i++) {
	    if (data[i].enclosure(h)) {
		Leaf l = ((AbstractNode) getChild(i)).findLeaf(h);
		if (l != null) {
		    return l;
		}
	    }
	}

	return null;
    }

    /**
     * Add the new HyperCube to all mbbs present in this node. Calculate the area difference and
     * choose the entry with the least enlargement. Based on that metric we choose the path that
     * leads to the leaf that will hold the new HyperCube.
     * [A. Guttman 'R-trees a dynamic index structure for spatial searching']
     *
     * @return The index of the branch of the path that leads to the Leaf where the new HyperCube 
     *         should be inserted.
     */
    private int findLeastEnlargement(HyperCube h) {
	double area = Double.POSITIVE_INFINITY;
	int sel = -1;

	for (int i = 0; i < usedSpace; i++) {
	    double enl = data[i].getUnionMbb(h).getArea() - data[i].getArea();
	    if (enl < area) {
		area = enl;
		sel = i;
	    } else if (enl == area) {
		sel = (data[sel].getArea() <= data[i].getArea()) ? sel : i;
	    }
	}
	return sel;
    }

    /**
     * Inserts a new node into the tree. If enought space is available
     * the insertion is straighforward. If not, a split must occur.
     *
     * @param  node The node that should be inserted.
     *
     * @return A Vector containing all segments that must be reinserted at the end.
     */
    protected Vector insert(AbstractNode node) {
	Vector res = new Vector();

	// remember that data array is one entry bigger than the capacity to
	// facilitate the splitting algorithm.
	if (usedSpace < data.length-1) {
	    data[usedSpace] = node.getNodeMbb();
	    branches[usedSpace] = node.pageNumber;
	    usedSpace++;
	    updateNodeMbb();
	    node.parent = pageNumber;
	    tree.file.writeNode(node);
	    tree.file.writeNode(this);
	    Index p = (Index) getParent();
	    if (p != null) {
		Vector v = p.adjustTree(this, null);
		res.addAll(v);
	    }
	} else {
	    Index[] a = splitIndex(node);
	    Index n = a[0];
	    Index nn = a[1];

	    if (isRoot()) {
		n.parent = 0;
		n.pageNumber = -1;
		nn.parent = 0;
		nn.pageNumber = -1;

		long p = tree.file.writeNode(n);
		for (int i = 0; i < n.usedSpace; i++) {
		    AbstractNode ch = (AbstractNode) n.getChild(i);
		    ch.parent = p;
		    tree.file.writeNode(ch);
		}

		p = tree.file.writeNode(nn);
		for (int i = 0; i < nn.usedSpace; i++) {
		    AbstractNode ch = (AbstractNode) nn.getChild(i);
		    ch.parent = p;
		    tree.file.writeNode(ch);
		}
		Index r = new Index(tree, SRTree.NIL, 0, level+1);
		r.addData(n.getNodeMbb(), n.pageNumber);
		r.addData(nn.getNodeMbb(), nn.pageNumber);
		r.updateNodeMbb();
		tree.file.writeNode(r);
		for (int i = 0; i < segmentSpace; i++) {
		    tree.insert(segments[i]);
		}
	    } else {
		n.pageNumber = pageNumber;
		n.parent = parent;
		nn.pageNumber = -1;
		nn.parent = parent;

		tree.file.writeNode(n);
		long j = tree.file.writeNode(nn);

		for (int i = 0; i < nn.usedSpace; i++) {
		    AbstractNode ch = (AbstractNode) nn.getChild(i);
		    ch.parent = j;
		    tree.file.writeNode(ch);
		}

		Index p = (Index) getParent();
		Vector v = p.adjustTree(n, nn);
		res.addAll(v);
	    }
	}

	return res;
    }

    protected long insertSegment(HyperCube h) {
	Vector reInsert = new Vector();
	int span = -1;

	// find spanned data entry.
	for (span = 0; span < usedSpace; span++) {
	    if (h.spansOver(data[span], 0)) {
		break;
	    }
	}

	// Check to see if segment extends beyond a boundary of this node.
	HyperCube p = getNodeMbb();
	HyperCube t = (HyperCube) h.clone();

	for (int i = 0; i < tree.getDimension(); i++) {
	    float l1 = p.getP1().getFloatCoordinate(i);
	    float h1 = p.getP2().getFloatCoordinate(i);
	    float l2 = t.getP1().getFloatCoordinate(i);
	    float h2 = t.getP2().getFloatCoordinate(i);

	    if (h2 > h1) {
		/**
		 *  -------
		 * |1      |
		 * |     --|--
		 * |    |2 |  |
		 * |     --|--
		 *  -------
		 */
		Point pt = (Point) t.getP1().clone();
		pt.data[i] = h1;
		HyperCube r = new HyperCube(pt, t.getP2());
		r.oid = t.oid;
		reInsert.addElement(r);
		
		pt = (Point) t.getP2().clone();
		pt.data[i] = h1;
		HyperCube s = new HyperCube(t.getP1(), pt);
		s.oid = t.oid;
		t = s;
	    } else if (l2 < l1) {
		/**
		 *     -------    
		 *    |1      |
		 *  --|--     |
		 * |2 |  |    |
		 *  --|--     |
		 *     ------- 
		 */
		Point pt = (Point) t.getP2().clone();
		pt.data[i] = l1;
		HyperCube r = new HyperCube(t.getP1(), pt);
		r.oid = t.oid;
		reInsert.addElement(r);
		
		pt = (Point) t.getP1().clone();
		pt.data[i] = l1;
		HyperCube s = new HyperCube(pt, t.getP2());
		s.oid = t.oid;
		t = s;
	    }
	}

	if (segmentSpace >= segments.length) {
	    System.out.println("Overflow of segment array. Ignoring entry.");
	    tree.overflow++;
	} else {
	    segments[segmentSpace] = t;
	    spanned[segmentSpace] = span;
	    segmentSpace++;
	}

	// Save changes.
	tree.file.writeNode(this);

	// insert remnant portion, if one was created.
	for (int i = 0; i < reInsert.size(); i++) {
	    tree.insert((HyperCube) reInsert.elementAt(i));
	}

	return pageNumber;
    }

    /**
     * Called by insert to adjust the parents of the node that was modified by the insertion. Stops
     * when it reaches the root.
     *
     * @param  n1  The child node that caused the need for adjustment.
     * @param  n2  If a split occured in the child node, this is the new node that must be inserted.
     *             Otherwise, this is null.
     * @return A vector containing all spanning segments that must be reinserted.
     */
    protected Vector adjustTree(AbstractNode n1, AbstractNode n2) {
	Vector res = new Vector();

	// find entry pointing to old node;
	for (int i = 0; i < usedSpace; i++) {
	    if (branches[i] == n1.pageNumber) {
		// update MBB.
		data[i] = n1.getNodeMbb();
		updateNodeMbb();
		// check spanning list of data entry to see if it is still spanned by those segments.
		for (int seg = 0; seg < segmentSpace; seg++) {
		    if (spanned[seg] == i && ! segments[seg].spansOver(data[i], 0)) {
			// store the hypercube, for reinsertion.
			res.addElement(segments[seg]);
			// remove the segment from the segment array.
			if (segmentSpace == 1) {
			    segments[0] = null;
			} else {
			    System.arraycopy(segments, seg+1, segments, seg, segmentSpace - seg - 1);
			}
			segmentSpace--;
		    }
		}
		break;
	    }
	}
	// save changes.
	tree.file.writeNode(this);

	// if a split has occured we must insert the new node, else we must continue adjusting the tree
	// until we hit the root.
	if (n2 != null) {
	    Vector v = insert(n2);
	    res.addAll(v);
	} else if (! isRoot()) {
	    Index p = (Index) getParent();
	    Vector v = p.adjustTree(this, null);
	    res.addAll(v);
	}

	return res;
    }

    private Index[] splitIndex(AbstractNode n) {
	int[][] group = null;

	group = quadraticSplit(n.getNodeMbb(), n.pageNumber);

	Index i1 = new Index(tree, parent, pageNumber, level);
	Index i2 = new Index(tree, parent, -1, level);

	int[] g1 = group[0];
	int[] g2 = group[1];

	// copy entries into correct Index node. Move all spanning segments
	// of a data entry with it.
	for (int i = 0; i < g1.length; i++) {
	    int originalIndex = g1[i];
	    int newIndex = i1.addData(data[originalIndex], branches[originalIndex]);

	    for (int seg = 0; seg < segmentSpace; seg++) {
	      if (spanned[seg] == originalIndex) {
		i1.segments[i1.segmentSpace] = segments[seg];
		i1.spanned[i1.segmentSpace] = newIndex;
		i1.segmentSpace++;
		if (segmentSpace == 1) {
		    segments[0] = null;
		} else {
		    System.arraycopy(segments, seg+1, segments, seg, segmentSpace - seg - 1);
		}
		segmentSpace--;
	      }
	    }
	}
	i1.updateNodeMbb();

	for (int i = 0; i < g2.length; i++) {
	    int originalIndex = g2[i];
	    int newIndex = i2.addData(data[originalIndex], branches[originalIndex]);
	    for (int seg = 0; seg < segmentSpace; seg++) {
	      if (spanned[seg] == originalIndex) {
		i2.segments[i2.segmentSpace] = segments[seg];
		i2.spanned[i2.segmentSpace] = newIndex;
		i2.segmentSpace++;
		if (segmentSpace == 1) {
		    segments[0] = null;
		} else {
		    System.arraycopy(segments, seg+1, segments, seg, segmentSpace - seg - 1);
		}
		segmentSpace--;
	      }
	    }
	}
	i2.updateNodeMbb();

	return new Index[] {i1, i2};
    }

    /**
     * Retrieves the <B>i-th</B> child node. Loads one page into main memory.
     *
     * @param  i The index of the child in the data array.
     * @return The i-th child.
     */
    public AbstractNode getChild(int i) {
	if (i < 0 || i >= usedSpace) {
	    throw new IndexOutOfBoundsException("" + i);
	}

	return tree.file.readNode(branches[i]);
    }

    public HyperCube[] getSegments() {
	HyperCube[] s = new HyperCube[segmentSpace];

	for (int i = 0; i < segmentSpace; i++) {
	    s[i] = (HyperCube) segments[i].clone();
	}

	return s;
    }

    public String toString() {
        String s = "< Page: " + pageNumber + ", Level: " + level + ", Used Space: " + usedSpace +
	  ", Segment Space: " + segmentSpace + ", Parent: " + parent + "\n  Mbb: " + mbb + "\n";

        s += " Data:\n";
	for (int i = 0; i < usedSpace; i++) {
	    s += "  " + (i+1) + ") " + data[i].toString() + " --> page: " + branches[i] + "\n";
	}
	s += " Segments:\n";
	for (int i = 0; i < segmentSpace; i++) {
	  s += "  " + (i+1) + ") " + segments[i].toString() + " --> spans: " + (spanned[i] + 1) + "\n";
	}
	s += ">\n";

	return s;
    }

  protected int getPageSize() {
      int data = (int) ((this.data.length - 1) * (8 * tree.getDimension() + 16));
      int segs = (int) (segments.length * (8 * tree.getDimension() + 12));

      //                  header         the mbb
      return data + segs  + 24 + (8 * tree.getDimension());
  }

} // Index
