package srtree;

import java.util.*;

/**
 * A Leaf node. Containts pointers to the real data.
 * <p>
 * Created: Tue May 18 13:04:08 1999
 * <p>
 * @author Hadjieleftheriou Marios
 * @version 1.002
 */
public class Leaf extends AbstractNode {
    
    protected Leaf(SRTree tree, long parent, long pageNumber) {
	// Leaf nodes belong by default to level 0.
	super(tree, parent, pageNumber, 0);

	int b = tree.getNodeCapacity() + 1;
	data = new HyperCube[b];
	branches = new long[b];
	usedSpace = 0;
    }

    protected Leaf(SRTree tree, long parent) {
	// Leaf nodes belong by default to level 0.
	this(tree, parent, -1);
    }

    protected AbstractNode chooseNode(HyperCube h) {
	return this;
    }

    protected Leaf findLeaf(HyperCube h) {
	for (int i = 0; i < usedSpace; i++) {
	    if (data[i].enclosure(h)) {
		return this;
	    }
	}

	return null;
    }

    /**
     * Inserts a new HyperCube into the Leaf. If there is space left, then the insertion is straightforward and
     * adjustTree must be called. Otherwise, a split must occur and two new leaves must be created.
     *
     * @param h The HyperCube that represents the MBB of the data.
     * @param page The page number where the real data resides.
     * @return The page number of the Leaf where the hypercube was inserted.
     */
    protected long insert(HyperCube h) {
	// remember that data array is one entry bigger than the capacity to
	// facilitate the splitting algorithm.
	if (usedSpace < data.length-1) {
	    data[usedSpace] = h;
	    branches[usedSpace] = h.oid;
	    usedSpace++;
	    updateNodeMbb();
	    tree.file.writeNode(this);
	    Index p = (Index) getParent();
	    if (p != null) {
		Vector v = p.adjustTree(this, null);
		for (int i = 0; i < v.size(); i++) {
		    tree.insert((HyperCube) v.elementAt(i));
		}
	    }
	    return pageNumber;
	} else {
	    Leaf[] a = splitLeaf(h);
	    Leaf l = a[0];
	    Leaf ll = a[1];

	    if (isRoot()) {
		// root is full, so we must split it. From now on root will be
		// an Index and not a Leaf.
		l.parent = 0;
		l.pageNumber = -1;
		ll.parent = 0;
		ll.pageNumber = -1;
		tree.file.writeNode(l);
		tree.file.writeNode(ll);
		// create the new root node. It belongs to pageNumber 0 and level 1.
		Index r = new Index(tree, SRTree.NIL, 0, 1);
		r.addData(l.getNodeMbb(), l.pageNumber);
		r.addData(ll.getNodeMbb(), ll.pageNumber);
		r.updateNodeMbb();
		tree.file.writeNode(r);
	    } else {
		// use old page number for left child, a new page number for the right child.
		l.pageNumber = pageNumber;
		ll.pageNumber = -1;
		tree.file.writeNode(l);
		tree.file.writeNode(ll);
		Index p = (Index) getParent();
		Vector v = p.adjustTree(l, ll);
		for (int i = 0; i < v.size(); i++) {
		    tree.insert((HyperCube) v.elementAt(i));
		}
	    }

	    // Find which is the parent Leaf of the inserted hypercube and return it's page number.
	    for (int i = 0; i < l.usedSpace; i++) {
		if (l.branches[i] == h.oid) {
		    return l.pageNumber;
		}
	    }

	    for (int i = 0; i < ll.usedSpace; i++) {
		if (ll.branches[i] == h.oid) {
		    return ll.pageNumber;
		}
	    }

	    return -1;
	}
    }

    private Leaf[] splitLeaf(HyperCube h) {
	int[][] group = null;

	group = quadraticSplit(h, h.oid);

	Leaf l = new Leaf(tree, parent);
	Leaf ll = new Leaf(tree, parent);

	int[] g1 = group[0];
	int[] g2 = group[1];

	for (int i = 0; i < g1.length; i++) {
	    l.addData(data[g1[i]], branches[g1[i]]);
	}
	l.updateNodeMbb();
	    
	for (int i = 0; i < g2.length; i++) {
	    ll.addData(data[g2[i]], branches[g2[i]]);
	}
	ll.updateNodeMbb();

	return new Leaf[] {l, ll};
    }

    /**
     * Returns the object identifier of the <i>i-th</i> data entry.
     *
     * @param  i The index of the child in the data array.
     * @return The pointer of the <i>i-th</i> child.
     */
    public long getObjectId(int i) {
	if (i < 0 || i >= usedSpace) {
	    throw new IndexOutOfBoundsException("" + i);
	}

	return branches[i];
    }

  protected int getPageSize() {
    return (data.length - 1) * (8 * tree.getDimension() + 8) + 20 + (8 * tree.getDimension());
  }

} // Leaf
