package srtree;

import java.io.*;
import java.util.*;

/**
 * A page file that stores all pages into Persistent storage. It uses a RandomAccessFile to store node data.
 * The format of the page file is as follows. First, a header is writen that stores important information
 * about the RTree.
 *
 * @author Hadjieleftheriou Marios
 * @version 1.001
 */
public class PersistentPageFile extends PageFile {
    /** Stores node data into Persistent storage. */
    private RandomAccessFile file;
    private String fileName;

    /**
     * Header size calculated using the following formula:
     *   headerSize = dimension + fillFactor + nodeCapacity + branchFactor
     */
    private int headerSize = 16;

    private int maxPageSize = 16384;

    public PersistentPageFile() {
	this(null);
    }

    public PersistentPageFile(String fileName) {
	try {
	    if (fileName == null) {
		File f = File.createTempFile("rtree", ".dat");
		this.fileName = f.getCanonicalPath();
		f.deleteOnExit();
		file = new RandomAccessFile(f, "rw");
	    } else {
		file = new RandomAccessFile(fileName, "rw");
		this.fileName = fileName;

		file.seek(0);
		byte[] header = new byte[headerSize];
		if (headerSize == file.read(header)) {
		    DataInputStream ds = new DataInputStream(new ByteArrayInputStream(header));
		    dimension = ds.readInt();
		    fillFactor = ds.readFloat();
		    nodeCapacity = ds.readInt();
		    branchFactor = ds.readFloat();
		}
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    protected void initialize(SRTree tree, int dimension, float fillFactor, int capacity, float branchFactor, int treeType) {
	super.initialize(tree, dimension, fillFactor, capacity, branchFactor, treeType);

	try {
	    // FIXME: needs Java 2 to compile.
	    file.setLength(0);
	    //
	    file.seek(0);
	    file.writeInt(dimension);
	    file.writeFloat(fillFactor);
	    file.writeInt(nodeCapacity);
	    file.writeFloat(branchFactor);
	} catch (IOException e) {
	    e.printStackTrace();
	}
    }

    protected void finalize() throws Throwable {
	try {
	    file.close();
	} catch (Exception e) {
	    e.printStackTrace();
	}

	super.finalize();
    }

    protected AbstractNode readNode(long offset) throws PageFaultError {
	reads++;

	if (offset < 0) {
	    throw new IllegalArgumentException("Offset cannot be negative.");
	}

	try {
	    file.seek(offset + headerSize);

	    byte[] b = new byte[maxPageSize];
	    int l = file.read(b);
	    if (-1 == l) {
		throw new PageFaultError("EOF found while trying to read page " + offset + ".");
	    }

	    DataInputStream ds = new DataInputStream(new ByteArrayInputStream(b));

	    long parent = ds.readLong();
	    int level = ds.readInt();
	    int usedSpace = ds.readInt();

	    AbstractNode n;
	    if (level != 0) {
		n = new Index(tree, parent, offset, level);
		((Index) n).segmentSpace = ds.readInt();
	    } else {
		n = new Leaf(tree, parent, offset);
	    }

	    float[] p1 = new float[dimension];
	    float[] p2 = new float[dimension];

	    for (int j = 0; j < dimension; j++) {
		p1[j] = ds.readFloat();
		p2[j] = ds.readFloat();
	    }

	    n.parent = parent;
	    n.level = level;
	    n.usedSpace = usedSpace;
	    n.mbb = new HyperCube(new Point(p1), new Point(p2));

	    for (int i = 0; i < usedSpace; i++) {
		for (int j = 0; j < dimension; j++) {
		    p1[j] = ds.readFloat();
		    p2[j] = ds.readFloat();
		}

		n.data[i] = new HyperCube(new Point(p1), new Point(p2));
		n.branches[i] = ds.readLong();
		n.data[i].oid = ds.readLong();
	    }
	    
	    if (n.level != 0) {
		Index in = (Index) n;
		for (int i = 0; i < in.segmentSpace; i++) {
		    for (int j = 0; j < dimension; j++) {
			p1[j] = ds.readFloat();
			p2[j] = ds.readFloat();
		    }

		    
		    in.segments[i] = new HyperCube(new Point(p1), new Point(p2));
		    in.spanned[i] = ds.readInt();
		    in.segments[i].oid = ds.readLong();
		}
	    }

	    return n;
	} catch (IOException e) {
	    e.printStackTrace();
	    return null;
	}
    }

    protected long writeNode(AbstractNode n) throws PageFaultError {
	writes++;

	if (n == null) {
	    throw new IllegalArgumentException("Node cannot be null.");
	}

	try {
	    long offset;

	    if (n.pageNumber < 0) {
		offset = file.length();
		n.pageNumber = offset - headerSize;
	    } else {
		offset = n.pageNumber + headerSize;
	    }

	    ByteArrayOutputStream bs = new ByteArrayOutputStream(maxPageSize);
	    DataOutputStream ds = new DataOutputStream(bs);

	    ds.writeLong(n.parent);
	    ds.writeInt(n.level);
	    ds.writeInt(n.usedSpace);

	    if (n.level != 0) {
		ds.writeInt(((Index) n).segmentSpace);
	    }

	    for (int j = 0; j < dimension; j++) {
		ds.writeFloat(n.mbb.getP1().getFloatCoordinate(j));
		ds.writeFloat(n.mbb.getP2().getFloatCoordinate(j));
	    }

	    for (int i = 0; i < n.usedSpace; i++) {
		for (int j = 0; j < dimension; j++) {
		    ds.writeFloat(n.data[i].getP1().getFloatCoordinate(j));
		    ds.writeFloat(n.data[i].getP2().getFloatCoordinate(j));
		}
		ds.writeLong(n.branches[i]);
		ds.writeLong(n.data[i].oid);
	    }

	    if (n.level != 0) {
		Index in = (Index) n;
		for (int i = 0; i < in.segmentSpace; i++) {
		    for (int j = 0; j < dimension; j++) {
			ds.writeFloat(in.segments[i].getP1().getFloatCoordinate(j));
			ds.writeFloat(in.segments[i].getP2().getFloatCoordinate(j));
		    }
			
		    ds.writeInt(in.spanned[i]);
		    ds.writeLong(in.segments[i].oid);
		}
	    }

	    int s = ds.size();
	    for (int i = 0; i < maxPageSize-s; i++) {
		ds.write(0);
	    }

	    ds.flush();
	    bs.flush();
	    file.seek(offset);
	    file.write(bs.toByteArray());

	    return n.pageNumber;
	} catch (IOException e) {
	    e.printStackTrace();
	    return -1;
	}
    }

    public void dumpFile() {
	try {
	    file.seek(0);
	    
	    byte[] b = new byte[headerSize];
	    int l = file.read(b);
	    if (-1 == l) {
		throw new PageFaultError("EOF found while trying to read file.");
	    }

	    DataInputStream ds = new DataInputStream(new ByteArrayInputStream(b));
	    System.out.print("Header:\n  Dimension: " + ds.readInt());
	    System.out.print(", Fill Factor: " + ds.readFloat());
	    System.out.print(", Capacity: " + ds.readInt());
	    System.out.println(", Branch Factor: " + ds.readFloat() + "\n");

	    while (file.getFilePointer() < file.length()) {
		long offset = file.getFilePointer();

		System.out.print("< Offset: " + (offset - headerSize));
		
		b = new byte[maxPageSize];
		l = file.read(b);
		if (-1 == l) {
		    throw new PageFaultError("EOF found while trying to read file.");
		}
		
		ds = new DataInputStream(new ByteArrayInputStream(b));
		
		long parent = ds.readLong();
		System.out.print(", Parent: " + parent);
		int level = ds.readInt();
		System.out.print(", Level: " + level);
		int usedSpace = ds.readInt();
		System.out.print(", Used space: " + usedSpace);
		
		int segmentSpace = -1;
		if (level != 0) {
		    segmentSpace = ds.readInt();
		    System.out.print(", Segment space: " + segmentSpace + "\n");
		}

		float[] p1 = new float[dimension];
		float[] p2 = new float[dimension];

		for (int j = 0; j < dimension; j++) {
		    p1[j] = ds.readFloat();
		    p2[j] = ds.readFloat();
		}
		System.out.println("  Mbb: " + new HyperCube(new Point(p1), new Point(p2)));
		
		System.out.println(" Data:");
		for (int i = 0; i < usedSpace; i++) {
		    for (int j = 0; j < dimension; j++) {
			p1[j] = ds.readFloat();
			p2[j] = ds.readFloat();
		    }
		    
		    HyperCube h = new HyperCube(new Point(p1), new Point(p2));
		    long o = ds.readLong();
		    h.oid = ds.readLong();
		    System.out.println("  " + (i + 1) + ") " + h + " --> offset: " + o);
		}

		if (level != 0) {
		    System.out.println(" Segments:");
		    for (int i = 0; i < segmentSpace; i++) {
			for (int j = 0; j < dimension; j++) {
			    p1[j] = ds.readFloat();
			    p2[j] = ds.readFloat();
			}

			HyperCube h = new HyperCube(new Point(p1), new Point(p2));
			int span = ds.readInt();
			long oid = ds.readLong();
			h.oid = oid;
			System.out.println("  " + (i + 1) + ") " + h + " --> spans: " + (span + 1));
		    }
		}
		System.out.println(">\n");
	    }
	    System.out.println("Reads: " + reads + ", Writes: " + writes);
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

} // PersistentPageFile

