package srtree;

import java.util.*;
import java.io.*;

/**
 * To create a new SRTree use the first two constructors. You must specify the dimension, the fill factor as
 * a float between 0 and 0.5 (0 to 50% capacity) and the variant of the RTree which is one of:
 * <ul>
 *  <li>RTREE_QUADRATIC</li>
 * </ul>
 * The first constructor creates by default a new memory resident page file. The second constructor takes
 * the page file as an argument. If the given page file is not empty, then all data are deleted.
 * <p>
 * The third constructor initializes the RTree from an already filled page file. Thus, you may store the
 * RTree into a persistent page file and recreate it again at any time.
 * <p>
 * Created: Tue May 18 12:57:35 1999
 * <p>
 * @author Hadjieleftheriou Marios
 * @version 1.003
 */
public class SRTree {
    public final String version = "1.000";
    public final String date = "January 30th 2000";

    /** Page file where data is stored. */
    protected PageFile file = null;

    /** static identifier used for the parent of the root node. */
    public static final long NIL = -1;

    /** Available RTree variants. */
    public static final int RTREE_QUADRATIC = 1;

    public int overflow = 0;

    /** 
     * Creates a memory resident SRTree with an empty root node, which is stored in page 0 and has a 
     * parent identifier equal to NIL. A new memory page file is created and used as a storage manager
     * by default.
     *
     * @param  fillFactor A percentage between 0 and 0.5. Used to calculate minimum number of entries
     *                    present in each node.
     * @param  capacity   The maximun number of entries that level 0 nodes can hold. Doubles at every level.
     * @param  branchFactor Used to calculate the number of branch entries present in its node.
     */
    public SRTree(int dimension, float fillFactor, int capacity, float branchFactor) {
	this(dimension, fillFactor, capacity, branchFactor, new MemoryPageFile());
    }

    /** 
     * Creates a new SRTree with an empty root node, which is stored in page 0 and has
     * a parent identifier equal to NIL. The given page file is used for storing the rtree nodes.
     * If the page file is not empty, than all entries will be overwritten.
     *
     * @param  fillFactor A percentage between 0 and 0.5. Used to calculate minimum number of entries
     *                    present in each node.
     * @param  capacity   The maximun number of entries that each node can hold.
     * @param  branchFactor Used to calculate the number of branch entries present in its node.
     * @param  file       The page file to use for storing the nodes of the rtree.
     */
    public SRTree(int dimension, float fillFactor, int capacity, float branchFactor, PageFile file) {
	if (dimension < 1) {
	    throw new IllegalArgumentException("Dimension must be larger than 1.");
	}

	if (fillFactor < 0 || fillFactor > 0.5) {
	    throw new IllegalArgumentException("Fill factor must be between 0 and 0.5.");
	}

	if (branchFactor < 0.5 || branchFactor > 1) {
	    throw new IllegalArgumentException("Branch factor must be between 0.5 and 1.");
	}

	if (capacity <= 1) {
	    throw new IllegalArgumentException("Capacity must be larger than 1.");
	}

	if (file.tree != null) {
	    throw new IllegalArgumentException("PageFile already in use by another rtree instance.");
	}

	file.initialize(this, dimension, fillFactor, capacity, branchFactor, RTREE_QUADRATIC);
	this.file = file;

	Leaf root = new Leaf(this, NIL, -1);
	file.writeNode(root);
    }

    /**
     * Creates an SRTree from an already initialized page file, probably stored into persistent storage.
     */
    public SRTree(PageFile file) {
	if (file.tree != null) {
	    throw new IllegalArgumentException("PageFile already in use by another rtree instance.");
	}

	if (file.dimension == -1) {
	    throw new IllegalArgumentException("PageFile is empty. Use some other RTree constructor.");
	}

	file.tree = this;
	this.file = file;
    }

    /**
     * Retruns the maximun capacity of each Node.
     */
    public int getNodeCapacity() {
	return file.nodeCapacity;
    }

    /** 
     * Returns the percentage between 0 and 0.5, used to calculate minimum number of entries
     * present in each node.
     */
    public float getFillFactor() {
	return file.fillFactor;
    }

    /** 
     * Returns the percentage between 0.5 and 1, used to calculate number of branches
     * present in each node.
     */
    public float getBranchFactor() {
	return file.branchFactor;
    }

    /** 
     * Returns the data dimension.
     */
    public int getDimension() {
	return file.dimension;
    }

    /** 
     * Returns the level of the root Node, which signifies the level of the whole tree. Loads one
     * page into main memory.
     */
    public int getTreeLevel() {
	return  file.readNode(0).getLevel();
    }

    /** 
     * Returns the RTree variant used. 
     */
    public int getTreeType() {
	return file.treeType;
    }

    /**
     * Inserts a HyperCube into the tree, pointing to the data stored at the given page number.
     *
     * @param h The hypercube to insert.
     * @return The offset of the page number of the Node where the hypercube was inserted (the parent of the
     *         data entry.)
     */
    public long insert(HyperCube h) {
	if (h  == null) {
	    throw new IllegalArgumentException("HyperCube cannot be null.");
	}

	if (h.getDimension() != getDimension()) {
	    throw new IllegalArgumentException("HyperCube dimension different than RTree dimension.");
	}

	AbstractNode root = file.readNode(0);

	AbstractNode n = root.chooseNode(h);
	
	if (n instanceof Index) {
	    return ((Index) n).insertSegment(h);
	}  else {
	    return ((Leaf) n).insert(h);
	}
    }

    /**
     * Returns a Vector containing all tree nodes from bottom to top, left to right.
     * CAUTION: If the tree is not memory resident, all nodes will be loaded into main memory.
     *
     * @param root The node from which the traverse should begin.
     * @return A Vector containing all Nodes in the correct order.
     */
    public Vector traverseByLevel(AbstractNode root) {
	if (root == null) {
	    throw new IllegalArgumentException("Node cannot be null.");
	}

	Vector ret = new Vector();
	Vector v = traversePostOrder(root);

	for (int i = 0; i <= getTreeLevel(); i++) {
	    Vector a = new Vector();
	    for (int j = 0; j < v.size(); j++) {
		Node n = (Node) v.elementAt(j);
		if (n.getLevel() == i) {
		    a.addElement(n);
		}
	    }
	    for (int j = 0; j < a.size(); j++) {
		ret.addElement(a.elementAt(j));
	    }
	}

	return ret;
    }

    /**
     * Returns an Enumeration containing all tree nodes from bottom to top, left to right.
     *
     * @return An Enumeration containing all Nodes in the correct order.
     */
    public Enumeration traverseByLevel() {
	class ByLevelEnum implements Enumeration {
	    // there is at least one node, the root node.
	    private boolean hasNext = true;

	    private Vector nodes;

	    private int index = 0;

	    public ByLevelEnum() {
		AbstractNode root = file.readNode(0);
		nodes = traverseByLevel(root);
	    }

	    public boolean hasMoreElements() {
		return hasNext;
	    }

	    public Object nextElement() {
		if (! hasNext) {
		    throw new NoSuchElementException("traverseByLevel");
		}

		Object n = nodes.elementAt(index);
		index++;
		if (index == nodes.size()) {
		    hasNext = false;
		}
		return n;
	    }
	};

	return new ByLevelEnum();
    }

    /**
     * Post order traverse of tree nodes. 
     * CAUTION: If the tree is not memory resident, all nodes will be loaded into main memory.
     *
     * @param root The node where the traversing should begin.
     * @return A Vector containing all tree nodes in the correct order.
     */
    public Vector traversePostOrder(AbstractNode root) {
	if (root == null) {
	    throw new IllegalArgumentException("Node cannot be null.");
	}

	Vector v = new Vector();
	v.addElement(root);

	if (root.isLeaf()) {
	} else {
	    for (int i = 0; i < root.usedSpace; i++) {
		Vector a = traversePostOrder(((Index) root).getChild(i));
		for (int j = 0; j < a.size(); j++) {
		    v.addElement(a.elementAt(j));
		}
	    }
	}
	return v;
    }

    /**
     * Post order traverse of all tree nodes, begging with root.
     * CAUTION: If the tree is not memory resident, all nodes will be loaded into main memory.
     *
     * @return An Enumeration containing all tree nodes in the correct order.
     */
    public Enumeration traversePostOrder() {
	class PostOrderEnum implements Enumeration {
	    private boolean hasNext = true;

	    private Vector nodes;

	    private int index = 0;

	    public PostOrderEnum() {
		AbstractNode root = file.readNode(0);
		nodes = traversePostOrder(root);
	    }

	    public boolean hasMoreElements() {
		return hasNext;
	    }

	    public Object nextElement() {
		if (! hasNext) {
		    throw new NoSuchElementException("traversePostOrder");
		}

		Object n = nodes.elementAt(index);
		index++;
		if (index == nodes.size()) {
		    hasNext = false;
		}
		return n;
	    }
	};

	return new PostOrderEnum();
    }

    /**
     * Pre order traverse of tree nodes.
     * CAUTION: If the tree is not memory resident, all nodes will be loaded into main memory.
     *
     * @param root The node where the traversing should begin.
     * @return A Vector containing all tree nodes in the correct order.
     */
    public Vector traversePreOrder(AbstractNode root) {
	if (root == null) {
	    throw new IllegalArgumentException("Node cannot be null.");
	}

	Vector v = new Vector();

	if (root.isLeaf()) {
	    v.addElement(root);
	} else {
	    for (int i = 0; i < root.usedSpace; i++) {
		Vector a = traversePreOrder(((Index) root).getChild(i));
		for (int j = 0; j < a.size(); j++) {
		    v.addElement(a.elementAt(j));
		}
	    }
	    v.addElement(root);
	}
	return v;
    }

    /**
     * Pre order traverse of all tree nodes, begging with root.
     * CAUTION: If the tree is not memory resident, all nodes will be loaded into main memory.
     *
     * @return An Enumeration containing all tree nodes in the correct order.
     */
    public Enumeration traversePreOrder() {
	class PreOrderEnum implements Enumeration {
	    private boolean hasNext = true;

	    private Vector nodes;

	    private int index = 0;

	    public PreOrderEnum() {
		AbstractNode root = file.readNode(0);
		nodes = traversePreOrder(root);
	    }

	    public boolean hasMoreElements() {
		return hasNext;
	    }

	    public Object nextElement() {
		if (! hasNext) {
		    throw new NoSuchElementException("traversePreOrder");
		}

		Object n = nodes.elementAt(index);
		index++;
		if (index == nodes.size()) {
		    hasNext = false;
		}
		return n;
	    }
	};

	return new PreOrderEnum();
    }
    
    /**
     * Returns a Vector with all nodes that intersect with the given HyperCube.
     * The nodes are returned in post order traversing
     *
     * @param h The given HyperCube that is tested for overlapping.
     * @param root The node where the search should begin.
     * @return A Vector containing the appropriate nodes in the correct order.
     */
    public Vector intersection(HyperCube h, AbstractNode root) {
	if (h == null || root == null) {
	    throw new IllegalArgumentException("Arguments cannot be null.");
	}

	if (h.getDimension() != file.dimension) {
	    throw new IllegalArgumentException("HyperCube dimension different than RTree dimension.");
	}

	Vector v = new Vector();
	
	if (root.getNodeMbb().intersection(h)) {
	    // v.addElement(root);

	    if (!root.isLeaf()) {
		for (int i = 0; i < root.usedSpace; i++) {
		    if (root.data[i].intersection(h)) {
			Vector a = intersection(h, ((Index) root).getChild(i));
			for (int j = 0; j < a.size(); j++) {
			    v.addElement(a.elementAt(j));
			}
		    }
		}
	    }
	}
	return v;
    }

    /**
     * Returns an Enumeration with all nodes present in the tree that intersect with the given
     * HyperCube. The nodes are returned in post order traversing
     *
     * @param h The given HyperCube that is tested for overlapping.
     * @return An Enumeration containing the appropriate nodes in the correct order.
     */
    public Enumeration intersection(HyperCube h) {
	class IntersectionEnum implements Enumeration {
	    private boolean hasNext = true;

	    private Vector nodes;

	    private int index = 0;

	    public IntersectionEnum(HyperCube hh) {
		nodes = intersection(hh, file.readNode(0));
		if (nodes.isEmpty()) {
		    hasNext = false;
		}
	    }

	    public boolean hasMoreElements() {
		return hasNext;
	    }
	    
	    public Object nextElement() {
		if (! hasNext) {
		    throw new NoSuchElementException("intersection");
		}

		Object c = nodes.elementAt(index);
		index++;
		if (index == nodes.size()) {
		    hasNext = false;
		}
		return c;
	    }
	};

	return new IntersectionEnum(h);
    }

    /**
     * Returns a Vector with all Hypercubes that completely contain HyperCube <B>h</B>.
     * The HyperCubes are returned in post order traversing, according to the Nodes where
     * they belong.
     *
     * @param h The given HyperCube.
     * @param root The node where the search should begin.
     * @return A Vector containing the appropriate nodes in the correct order.
     */
    public Vector enclosure(HyperCube h, AbstractNode root) {
	if (h == null || root == null) throw new
	    IllegalArgumentException("Arguments cannot be null.");

	if (h.getDimension() != file.dimension) throw new
	    IllegalArgumentException("HyperCube dimension different than RTree dimension.");

	Vector v = new Vector();

	if (root.getNodeMbb().enclosure(h)) {
	    v.addElement(root);

	    if (!root.isLeaf()) {
		for (int i = 0; i < root.usedSpace; i++) {
		    if (root.data[i].enclosure(h)) {
			Vector a = enclosure(h, ((Index) root).getChild(i));
			for (int j = 0; j < a.size(); j++) {
			    v.addElement(a.elementAt(j));
			}
		    }
		}
	    }
	}
	return v;
    }

    /**
     * Returns an Enumeration with all Hypercubes present in the tree that contain the given
     * HyperCube. The HyperCubes are returned in post order traversing, according to the Nodes where
     * they belong.
     *
     * @param h The given HyperCube.
     * @param root The node where the search should begin.
     * @return An Enumeration containing the appropriate HyperCubes in the correct order.
     */
    public Enumeration enclosure(HyperCube h) {
	class ContainEnum implements Enumeration {
	    private boolean hasNext = true;

	    private Vector cubes;

	    private int index = 0;

	    public ContainEnum(HyperCube hh) {
		cubes = enclosure(hh, file.readNode(0));
		if (cubes.isEmpty()) {
		    hasNext = false;
		}
	    }

	    public boolean hasMoreElements() {
		return hasNext;
	    }
	    
	    public Object nextElement() {
		if (!hasNext) throw new
		    NoSuchElementException("enclosure");

		Object c = cubes.elementAt(index);
		index++;
		if (index == cubes.size()) {
		    hasNext = false;
		}
		return c;
	    }
	};

	return new ContainEnum(h);
    }

    /**
     * Returns a Vector with all Hypercubes that completely contain point <B>p</B>.
     * The HyperCubes are returned in post order traversing, according to the Nodes where
     * they belong.
     *
     * @param p The given point.
     * @param root The node where the search should begin.
     * @return A Vector containing the appropriate HyperCubes in the correct order.
     */
    public Vector enclosure(Point p, AbstractNode root) {
	return enclosure(new HyperCube(p, p), root);
    }

    /**
     * Returns an Enumeration with all Hypercubes present in the tree that contain the given
     * point. The HyperCubes are returned in post order traversing, according to the Nodes where
     * they belong.
     *
     * @param p The query point.
     * @param root The node where the search should begin.
     * @return An Enumeration containing the appropriate HyperCubes in the correct order.
     */
    public Enumeration enclosure(Point p) {
	return enclosure(new HyperCube(p, p));
    }

    /**
     * Returns the nearest HyperCube to the given point.
     * [King Lum Cheung and Ada Wai-chee Fu: Enhanced Nearest Neighbor Search on the R-Tree]
     *
     * @param p The query point.
     * @return A vector containing all the nodes lying in the search path until the nearest hypercube
     *         is found. Elements are instances of AbstractNode and Data classes. The last Data instance
     *         in the vector is the answer to the query.
     */
    public Vector nearestNeighbor(Point p) {
	return nearestNeighborSearch(file.readNode(0), p, Float.POSITIVE_INFINITY);
    }

    /**
     * Used for nearest neighbor recursive search into the RTree structure. <B>n</B> is the current
     * node of the active branch list, searched. <B>p</B> is the query point. <B>nearest</B> is the 
     * distance of <B>p</B> from current nearest hypercube <B>h</B>.
     */
    protected Vector nearestNeighborSearch(AbstractNode n, Point p, float nearest) {
	Vector ret = new Vector();
	HyperCube h;

	if (n.isLeaf()) {
	    for (int i = 0; i < n.usedSpace; i++) {
		float dist = n.data[i].getMinDist(p);
		if (dist < nearest) {
		    h = n.data[i];
		    nearest = dist;
		    ret.addElement(new Data(h, n.branches[i], n.pageNumber, i));
		}
	    }
	    return ret;
	} else {
	    // generate Active Branch List.
	    class ABLNode {
		AbstractNode node;
		float minDist;

		public ABLNode(AbstractNode node, float minDist) {
		    this.node = node;
		    this.minDist = minDist;
		}
	    }

	    ABLNode[] abl = new ABLNode[n.usedSpace];

	    for (int i = 0; i < n.usedSpace; i++) {
		AbstractNode ch = ((Index) n).getChild(i);
		abl[i] = new ABLNode(ch , ch.getNodeMbb().getMinDist(p));
	    }

	    // sort ABL in ascending order of MINDIST from the query point.
	    Sort.mergeSort(
		abl,
		new Comparator() {
			public int compare(Object o1, Object o2) {
			    float f = ((ABLNode) o1).minDist - 
				        ((ABLNode) o2).minDist;

			    // do not round the float here. It is wrong!
			    if (f > 0) return 1;
			    else if (f < 0) return -1;
			    else  return 0;
			}
		}
	    );

	    // traverse all ABL nodes and prune irrelevant nodes according to the MINDIST heuristic.
	    for (int i = 0; i < abl.length; i++) {
		if (abl[i].minDist <= nearest) {
		    // add node in the results vector, if it complies to the MINDIST heuristic.
		    ret.addElement(abl[i].node);
		    // recursively continue the search.
		    Vector v = nearestNeighborSearch(abl[i].node, p, nearest);

		    // find the new nearest distance and add all the nodes accessed, into the current 
		    // results vector.
		    try {
			Object o = v.lastElement();
			if (o instanceof AbstractNode) {
			    for (int j = 0; j < v.size(); j++) {
				ret.addElement(v.elementAt(j));
			    }
			    AbstractNode an = (AbstractNode) o;
			    h = (HyperCube) an.getNodeMbb();
			    nearest = h.getMinDist(p);
			} else if (o instanceof Data) {
			    // if the current node searched was a leaf, than the resulting set definetly
			    // contains just one HyperCube, the nearest to the query point.
			    h = ((Data) o).mbb;
			    nearest = h.getMinDist(p);
			    ret.addElement(o);
			}
		    } catch (NoSuchElementException e) {
			// no nearest node or hypercube was found from this recursion step. Leave nearest
			// distance intact.
		    }
		}
	    }

	    return ret;
	}
    }

    public static void main(String[] args) {
	CachedPersistentPageFile file = new CachedPersistentPageFile("../srtree-data/max-4.tree", 10);
	//CachedPersistentPageFile file = new CachedPersistentPageFile("c:/temp/rtree.dat", 3);
	//MemoryPageFile file = new MemoryPageFile();
	SRTree r = new SRTree(3, 0.5f, 50, 2.0f/3.0f, file);
	//SRTree r = new SRTree(file);

	int j = 0;
	long start = System.currentTimeMillis();

	try {
	    LineNumberReader lr = new LineNumberReader(new FileReader(new File("../srtree-data/max-4.dat")));
	    String s;

	    while ((s = lr.readLine()) != null) {
	    	StringTokenizer st = new StringTokenizer(s);
		String l = st.nextToken();
		int i = new Integer(st.nextToken()).intValue();
		int id = new Integer(st.nextToken()).intValue();
		float x1 = new Float(st.nextToken()).floatValue();
		float x2 = new Float(st.nextToken()).floatValue();
		float y1 = new Float(st.nextToken()).floatValue();
		float y2 = new Float(st.nextToken()).floatValue();
		float t1 = new Float(st.nextToken()).floatValue();
		float t2 = new Float(st.nextToken()).floatValue();

		if (i == 1) {
		    j++;
		    Point p1 = new Point(new float[] {t1, x1, y1});
		    Point p2 = new Point(new float[] {t2, x2, y2});
		    HyperCube h = new HyperCube(p1, p2);
		    h.oid = id;
		    r.insert(h);
		    //		    r.intersection(h);
		    //		    file.clearCache();
		}
	    }
	} catch (Exception e) {
	    e.printStackTrace();
	}

	long end = System.currentTimeMillis();

	file.dumpFile();

	System.out.println("Read: " + j + ", Overflows: " + r.overflow);
	System.out.println("Reads: " + file.reads + ", Writes: " + file.writes);
	//	System.out.println("Average Reads: " + (file.reads / j) + ", Average Writes: " + (file.writes / j));
	System.out.println("Minutes: " + (end - start) / 60000);
    }

} // SRTree
